import click
import os
from mymodule import mkdirs

plugin_folder = os.path.join(os.path.dirname(__file__), 'plugins')
mkdirs(plugin_folder)

def list_plugins(rv=[]):
    for filename in os.listdir(plugin_folder):
        if filename.endswith('.py'):
            temp = filename[:-3]
            if temp not in rv:
                rv.append(temp)
    return rv

class MyCLI(click.MultiCommand):
    """
    lazy load of sub commands from commands folder. 
    """
    commandList = {}
    def list_commands(self, ctx):
        rv = list(self.commandList.keys())
        rv = list_plugins(rv)
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        if name in self.commandList:
            return self.commandList[name]
        fn = os.path.join(plugin_folder, name + '.py')
        ns = {'__file__':fn}
        if os.path.isfile(fn):
            with open(fn) as f:
                # code = compile(f.read(), fn, 'exec')
                exec(f.read(), ns, ns)
            return ns['cli']
        else:
            return None

    def add_command(self,f):
        self.commandList[f.__name__] = f

    def add_command(self, cmd, name=None):
        """Registers another :class:`Command` with this group.  If the name
        is not provided, the name of the command is used.
        """
        name = name or cmd.name
        if name is None:
            raise TypeError("Command has no name.")
        if name in self.list_commands(1):
            raise RuntimeError(f'Command {name} already exist.')
        self.commandList[name] = cmd

    def command(self, *args, **kwargs):
        """A shortcut decorator for declaring and attaching a command to
        the group.  This takes the same arguments as :func:`command` but
        immediately registers the created command with this instance by
        calling into :meth:`add_command`.
        """
        command = click.command

        def decorator(f):
            cmd = command(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd
        return decorator


@click.group(cls=MyCLI)
def _plugins():
    pass 


@_plugins.command()
@click.option('--show-folder',"-sf",default=False, is_flag=True,help="Open plugins folder.")
@click.option('--install',"-i","file",default=None,type=click.Path(exists=True,dir_okay=False), help="Install to plugins folder.") #
@click.pass_context
def plugins(ctx,show_folder,file):
    """
    Manage plugins. Show folder or install plugin.
    Call directly will show a list of all currently available plugins.
    """
    if show_folder:
        import subprocess
        click.echo(f"Plugins stored in: {plugin_folder}")
        subprocess.run(f"cd {plugin_folder}\nopen .\n",shell=True)
        return 
    if file:
        from shutil import copyfile
        from cli.ok import menu
        cmds = menu.list_commands(ctx)
        click.echo(f"Current commands: {cmds}")
        name=click.prompt("Enter a different name for your plugin")
        if name in cmds:
            click.echo(f'!Failed. <{name}> is alread in use.')
            ctx.exit()
        copyfile(file,os.path.join(plugin_folder,f"{name}.py"))
        click.echo(f"Command <{name}> installed to plugins folder.")
        return

    # display installed plugin list 
    rv = list_plugins()
    click.echo(f'Currently installed plugins: {rv}')


#automated error checking for RNAstructure python interface
from __future__ import print_function
import inspect
from functools import wraps
from collections import defaultdict
debug = False
class StructureError(Exception): pass
class RNAstructureInternalError(Exception):pass
lookup_exceptions = defaultdict(lambda:RuntimeError,
        {   1:IOError,
            2:IOError,
            3:IndexError,
            4:IndexError,
            5:EnvironmentError,
            6:StructureError,
            7:StructureError,
            8:StructureError,
            9:StructureError,
            10:ValueError,
            11:ValueError,
            12:ValueError,
            13:IOError,
            14:RNAstructureInternalError,
            15:ValueError,
            16:ValueError,
            17:ValueError,
            18:ValueError,
            19:ValueError,
            20:ValueError,
            21:RNAstructureInternalError,
            22:RNAstructureInternalError,
            23:ValueError,
            24:ValueError,
            25:ValueError,
            26:ValueError
            })

def check_for_errors(method):
    @wraps(method)
    def RNAstructure_error_checker(self,*args,**kwargs):
        if debug: print ("checking for errors in %s" % method.__name__)
        ret = method(self,*args,**kwargs)
        error = self.GetErrorCode()
        self.ResetError()
        if error != 0:
            raise lookup_exceptions[error]("Error in %s: " % method.__name__ +
                                self.GetErrorMessage(error))
        return ret
    return RNAstructure_error_checker

def check_for_init_errors(method):
    @wraps(method)
    def RNAstructure_error_checker(self,*args):
        if debug: print ("checking for errors in %s" % method.__name__)
        ret = method(self,*args)
        error = self.GetErrorCode()
        if error != 0:
            raise RuntimeError("Error in call to %s.%s: " % (self.__name__,method.__name__) +
                                self.GetErrorMessage(error))
        return ret
    return RNAstructure_error_checker

def is_init(method):
    result =  inspect.ismethod(method) and method.__name__=="__init__"
    if inspect.ismethod(method):
        pass
    return result

def not_excluded(method):
    excluded = ["__repr__","__setattr__","__getattr__","__str__","__init__","<lambda>","swig_repr",
                "GetErrorCode","GetErrorMessage","GetErrorMessageString","ResetError","fromFile","fromString"]
    result = (inspect.ismethod(method) or inspect.isfunction(method)) and method.__name__ not in excluded
    if inspect.ismethod(method):
        if debug: print ("checking if", method.__name__ , "should be excluded: ",result)
    return result

def decorate_methods(decorator,methodtype):
    def decorate(cls):
        for attr in inspect.getmembers(cls, methodtype):
            if debug: print ("decorating %s!" % attr[0])
            setattr(cls, attr[0], decorator(getattr(cls, attr[0])))
        return cls
    return decorate



import re,json,os
from colorama import init, Fore,Back,Style
from textwrap import TextWrapper
from terminaltables import SingleTable
import subprocess 
from mymodule import mkdirs

WINDOWSIZE=[int(i) for i in subprocess.run('stty size', shell=True, stdout=subprocess.PIPE,encoding='utf-8').stdout.split()] 

TermRow, TermCol =  WINDOWSIZE

class ColorText():
    """
    Fore: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
    Back: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
    Style: DIM, NORMAL, BRIGHT, RESET_ALL
    """
    _color ={"K":"BLACK","R":"RED","Y":"YELLOW","G":"GREEN","B":'BLUE',"M":"MAGENTA","C":"CYAN","W":"WHITE",}
    _style={"D":"DIM","N":"NORMAL","B":"BRIGHT"}
    _mapping = {"f":_color,"s":_style,"b":_color}
    _type={'f':Fore,'b':Back,"s":Style}
    def __init__(self,mapping={"[]":"fC","{}":"bKfY",('<i>','</i>'):"fKbW", ("<g>","</g>") : "fG", ("<b>","</b>") : "fB", 
                ("<r>","</r>") : "fR",("<m>","</m>") : "fM",("<y>","</y>") : "fY", ("<a>","</a>"):"bR",}):
        """
        mapping: { pattern : color }
        pattern: "[]"/"{}"/(<p>,</p>)
        color:"fBbWsD"
        """
        self.color = {}
        ptns = []
        for k,i in mapping.items():
            color = ""
            for (s,c) in zip(i[::2],i[1::2]):
                color+=getattr(self._type[s],self._mapping[s][c])
            ptns.append((f"{re.escape(k[0])}(?P<{i}>.*?){re.escape(k[1])}"))
            self.color[i]=color
        self.pattern = re.compile("|".join(ptns),flags=re.DOTALL)

    def format(self,text):
        return self(text)

    def __call__(self,text,):
        def repl(x):
            mk = [k for k,i in x.groupdict().items() if i!=None][0]
            ct = x.groupdict()[mk]
            return self.color[mk]+ct+Style.RESET_ALL
        return self.pattern.sub(repl,text)


class TableDisplay():
    def __init__(self,color=ColorText()):
        self.col = (TermCol)
        if isinstance(color,dict):
            self.color = ColorText(color)
        else:
            self.color=color

    def format(self,title="",text=[]):
        """
        text = [ [ row 1 [ line1, line2 ] ], [row 2 [ lines]] ]
        if text is a string, then convert to single row, but respect the \n
        """
        if isinstance(text,str):
            text = [[i for i in text.split('\n')]]
        tw = TextWrapper(fix_sentence_endings=True,width=self.col-4)
        table_data = [ [self.color("\n".join(tw.fill(j,) for j in i))] for i in text]
        table = SingleTable(table_data,self.color(title))
        table.inner_row_border = True
        return table.table

    def __call__(self,title="",text=[]):
        return self.format(title,text)



class Config():
    """
    Class to store and read saved configure data.
    """
    folder = os.path.join(os.path.dirname(__file__), 'conf')
    def __init__(self,name):
        self.name = name 
        self.create()

    def create(self):
        mkdirs(self.folder)
        if not os.path.isfile(self.path):
            _=open(self.path,'wt')
            _.write("{}")
            _.close()
        
    @property
    def path(self):
        return os.path.join(self.folder,self.name+'.json')

    def readData(self):
        with open(self.path,'rt') as f:
            return json.load(f)

    def saveData(self,data,):
        with open(self.path,'wt') as f:
            json.dump(data,f,indent=2,)

    @staticmethod
    def list_config():
        fd = Config.folder
        mkdirs(fd)
        res = []
        for fn in os.listdir(fd):
            if fn.endswith('.json'):
                res.append(fn[:-5])
        return res


"""
How to use NUPACK - Hui 5/8
from mymodule import mymodule as NPK

# To predict mfe structure:
# pseudo: pseudo knot , sodium and magnesium in M,
NPK.mfe(['GGGGATACCCC'],dangles='some',T=37,material='dna',pseudo=False,sodium=0.05,magnesium=0.005)
>>> [('(((....))).', '-2.685')]

"""
from ._NUPACK import *



import click
import subprocess as sub
from cli.utils import TableDisplay, Config
import os 

config = Config('folders')


td = TableDisplay()

def displayMenu(data):
    data = data.get('favorites',{})
    dis = []
    for k in sorted(data.keys()):
        dis.append(f" >><g> [ {k} ] </g><< {data[k]}  ")
    click.echo(td(title= "[>> Favorite Folders <<]", text=[dis]))


def mark_fav(ctx,param,value):
    if not value or ctx.resilient_parsing:
        return
    cwd = os.getcwd()
    key = click.prompt(td(text=f'Enter name for favorite folder\n@ [{cwd}]')+'\n')
   
    data = config.readData()
    if data.get('favorites',None) is None:
        data['favorites'] = {}

    todis = []
    for k in key.split():
        data['favorites'][k]=cwd
        todis.append(f" >><g> [ {k} ] </g><< {cwd} ")

    config.saveData(data)
    click.echo(td(title=">>> <g>Favorite folder Saved</g> <<<",
                text=[todis]))
    ctx.exit()

def drop_fav(ctx,param,value):
    if not value or ctx.resilient_parsing:
        return
    data = config.readData()

    fav = data.get('favorites',{})
    displayMenu(data)
    key = click.prompt('Choose folder(s) to Delete',default="",show_default=False)
    dis = []
    if not key: ctx.abort()
    for k in key.split():
        if k in fav:
            dis.append(f"Deleted >><g> [ {k} ] </g><< {fav.pop(k)}")
        else:
            dis.append(f">> <a>{k}</a> << is not in favorites.")
    click.echo(td(text=[dis]))
    config.saveData(data)
    ctx.exit()

def config_shell(ctx,p,value):
    if not value or ctx.resilient_parsing:
        return
    data = config.readData()
    
    cs = data.get('shell','zsh')

    data['shell']  = click.prompt('Choose which shell to use', 
        type=click.Choice(['zsh', 'bash'],case_sensitive=False), 
        default='zsh',show_default=True )

    config.saveData(data)
    click.echo(f'Changed shell to {data["shell"]}.')
    ctx.exit()


@click.command() 
@click.option('--mark','-m',callback=mark_fav,is_flag=True,is_eager=True,
            expose_value=False,help="Mark current dir to favorite.")
@click.option('--drop','-d',callback=drop_fav,is_flag=True,is_eager=True,
            expose_value=False,help="Remove favorite folder.")
@click.option('--config','-c',callback=config_shell,is_flag=True,is_eager=True,
            expose_value=False,help="Configure default shell. Default zsh")
@click.argument('path',nargs=1,default=None,required=False)
@click.pass_context
def folder(ctx,path):
    """
    cd to favorite folders by $ fd \n
    Default shell is zsh, if need to use bash, use fd -c to config.
    """
    data = config.readData()
    fav=data.get('favorites',{})

    if path is None:
        displayMenu(data)
        path = click.prompt('Choose folder',default="",show_default=False)
        if not path: ctx.abort()
    
    if path not in fav:
        click.echo(td(text=f"'<r>{path}</r>' isn't in your favorites."))
        ctx.exit()

    if os.path.isdir(fav[path]):
        sub.run(f"cd '{fav[path]}'; exec {data.get('shell','zsh')} ",shell=True,)
    else:
        click.echo(td(text=f"Folder '<y>{fav[path]}</y>' doesn't exist."))





"""
RNAstructure.py by Michael Sloma (last update: 3/11/2015)

This is the Python interface to the RNAstructure C++ library
It exports the following classes:
    RNA
    Dynalign_object
    Multilign_object
    Oligowalk_object
    ProbScan

This is mostly a thin wrapper around the C++ classes that has been
automatically generated by SWIG, but it adds some features including
automatic error checking, pretty printing, and some convenience functions.
For object construction,class methods fromString and fromFile have been added
to each class, which are more user-friendly than the C++ constructors.

For further documentation and examples of usage, see the online documentation
at http://rna.urmc.edu


20190713
Hui isolated files from wrapped package and make it standalone.
"""

from __future__ import print_function

try:
    from builtins import object
except ImportError:
    from __builtin__ import object

from . import RNAstructure_wrap
from . import Error_handling as e
import re

#these are named this way to cooperate with python name manglign
filecodes = {"ct":1,"seq":2,"pfs":3,"sav":4}
polymers = {"rna":True,"RNA":True,"Rna":True,"dna":False,"DNA":False,"Dna":False}
invalid_nucs = re.compile(r'[^AUCGTaucgtxX]').search

def sequence_valid(sequence):
    if bool(invalid_nucs(sequence)):
        raise RuntimeError( "unrecognized nucleotides in sequence" )

def polymer_valid(polymer):
    if polymer not in polymers:
        raise RuntimeError( "unknown strand type '%s'" % polymer )

def filetype_valid(filetype):
    if filetype not in filecodes:
        raise RuntimeError( "unknown file type '%s'" % filetype )

def interpret_filetype(filepath):
    """Attempts to automatically interpret the file type of a seq or ct file \
raises an exception if you hand it anything else"""
    buf = open(filepath)
    firstline = buf.readline()
    secondline = buf.readline()
    if re.match(r'\s*[0-9]+\s+\D(\s+[0-9]+){4,4}',secondline):
        return "ct"
    elif re.match(r"^;",firstline.strip()):
        return "seq"
    else:
        raise IOError( "could not automatically determine sequence file type for RNA initialization: \
please provide file type" )

@e.decorate_methods(e.check_for_errors,e.not_excluded)
class RNA(RNAstructure_wrap.RNA):
    """The RNA class provides an entry point for all the single sequence operations of RNAstructure."""
    @e.check_for_errors
    def __init__(self,*args):
        """Initializes an RNA object using a thin wrapper around the C++ constructor (see
        RNAstructure docs for more details.) The class methods RNA.fromFile and
        RNA.fromString provide a better way to initialize RNA objects.
        """
        RNAstructure_wrap.RNA.__init__(self,*args)

    def __iter__(self):
        return self.iterNucs()

    def __len__(self):
        return self.GetSequenceLength()

    def __repr__(self):
        seq = "".join(nuc for nuc in self)
        return "RNA: %s%s" % (seq if len(self)<20 else seq[:15],
                              "" if len(self)<20 else "... %s"%seq[-4:])
    def __str__(self):
        return "".join(nuc for nuc in self)

    @classmethod
    def fromFile(cls,seqfile,filetype=None,backbone="rna"):
        """Create an RNA from an input file.
        Args:
            seqfile(str): A path to the location of the input file
            filetype(str): The format of the input file. Supported options are "seq",
                "ct","sav" (save file from the Fold program), and "pfs" (save file
                from the partition program). If no file type is provided, this function
                will attempt to infer the filetype automatically, and raise an IOError
                if the file cannot be understood.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        if filetype is None:
            filetype = interpret_filetype(seqfile)
        filetype_valid(filetype)
        polymer_valid(backbone)
        return cls(seqfile,filecodes[filetype],polymers[backbone])

    @classmethod
    def fromString(cls,sequence,backbone = "rna"):
        """Create an RNA object from a sequence
        Args:
            sequence(str): The sequence for the RNA. Input sequence should contain
                A,C,G,T,U,a,c,g,t,u,x,X. Capitalization makes no difference. If the
                sequence is an RNA, T is treated as U.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        polymer_valid(backbone)
        sequence_valid(sequence)
        return cls(sequence,polymers[backbone])

    def iterNucs(self):
        """iterate over nucleotide identities.
        note that this is has the same behavior as iterating directly
        on the RNA object."""
        return NucIterator(self)

    def iterIndices(self):
        """iterate over 1-indexed nucleotide indices"""
        return range(1,self.GetSequenceLength()+1)

@e.decorate_methods(e.check_for_errors,e.not_excluded)
class HybridRNA(RNAstructure_wrap.HybridRNA):
    @e.check_for_errors
    def __init__(self,*args):
        """Initializes a HybridRNA using a thin wrapper around the C++ constructor (see
        RNAstructure docs for more details.) The class methods HybridRNA.fromFile and
        HybridRNA.fromString provide a better way to initialize HybridRNA objects.
        """
        RNAstructure_wrap.HybridRNA.__init__(self,*args)

    @classmethod
    def fromFile(cls,seqfile1,seqfile2,filetype=None,backbone='rna'):
        """Create a HybridRNA from an input file.
        Args:
            seqfile1(str): A path to the location of the first input file
            seqfile2(str): A path to the location of the second input file
            filetype(str): The format of the input files. Supported options are "seq",
                "ct","sav" (save file from the Fold program), and "pfs" (save file
                from the partition program). If no file type is provided, this function
                will attempt to infer the filetype automatically, and raise an IOError
                if the file cannot be understood.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        if filetype is None:
            filetype1 = interpret_filetype(seqfile1)
            filetype2 = interpret_filetype(seqfile1)
        else:
            filetype_valid(filetype)
            filetype1 = filetype2 = filetype
        polymer_valid(backbone)
        return cls(seqfile1,filecodes[filetype1],seqfile2,
                filecodes[filetype2],polymers[backbone])

    @classmethod
    def fromString(cls,sequence1,sequence2,backbone='rna'):
        """Create a HybridRNA object from sequences
        Args:
            sequence1(str): The sequence of the first RNA. Input sequence should contain
                A,C,G,T,U,a,c,g,t,u,x,X. Capitalization makes no difference. If the
                sequence is an RNA, T is treated as U.
            sequence2(str): The sequence of the first RNA. Input sequence should contain
                A,C,G,T,U,a,c,g,t,u,x,X. Capitalization makes no difference. If the
                sequence is an RNA, T is treated as U.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        sequence_valid(sequence1)
        sequence_valid(sequence2)
        polymer_valid(backbone)
        return cls(sequence1,sequence2,polymers[backbone])


@e.decorate_methods(e.check_for_errors,e.not_excluded)
class Dynalign_object(RNAstructure_wrap.Dynalign_object):
    """Calculates the optimal structure common to RNAs by simultaneously aligning and folding"""
    @e.check_for_errors
    def __init__(self,*args):
        """Initializes a Dynalign_object using a thin wrapper around the C++ constructor (see
        RNAstructure docs for more details.) The class methods Dynalign_object.fromFile and
        Dynalign_object.fromString provide a better way to initialize RNA objects.
        """
        RNAstructure_wrap.Dynalign_object.__init__(self,*args)

    @classmethod
    def fromFile(cls,seqfile1,seqfile2,filetype=None,backbone='rna'):
        """Create a Dynalign_object from an input file.
        Args:
            seqfile1(str): A path to the location of the first input file
            seqfile2(str): A path to the location of the second input file
            filetype(str): The format of the input files. Supported options are "seq",
                "ct","sav" (save file from the Fold program), and "pfs" (save file
                from the partition program). If no file type is provided, this function
                will attempt to infer the filetype automatically, and raise an IOError
                if the file cannot be understood.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        if filetype is None:
            filetype1 = interpret_filetype(seqfile1)
            filetype2 = interpret_filetype(seqfile1)
        else:
            filetype_valid(filetype)
            filetype1 = filetype2 = filetype
        polymer_valid(backbone)
        return cls(seqfile1,filecodes[filetype1],seqfile2,
                filecodes[filetype2],polymers[backbone])

    @classmethod
    def fromString(cls,sequence1,sequence2,backbone='rna'):
        """Create a Dynalign_object object from sequences
        Args:
            sequence1(str): The sequence of the first RNA. Input sequence should contain
                A,C,G,T,U,a,c,g,t,u,x,X. Capitalization makes no difference. If the
                sequence is an RNA, T is treated as U.
            sequence2(str): The sequence of the first RNA. Input sequence should contain
                A,C,G,T,U,a,c,g,t,u,x,X. Capitalization makes no difference. If the
                sequence is an RNA, T is treated as U.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        sequence_valid(sequence1)
        sequence_valid(sequence2)
        polymer_valid(backbone)
        return cls(sequence1,sequence2,polymers[backbone])

@e.decorate_methods(e.check_for_errors,e.not_excluded)
class Multilign_object(RNAstructure_wrap.Multilign_object):
    """This class allows calculation of the common structure for two or more sequences
    by repeated pairwise Dynalign calculations."""
    @e.check_for_errors
    def __init__(self,*args):
        """Initializes a Multilign_object using a thin wrapper around the C++ constructor (see
        RNAstructure docs for more details.) The class method Multilign_object.fromFile
        provides a better way to initialize Multilign_objects.
        """
        RNAstructure_wrap.Multilign_object.__init__(self,*args)

    @classmethod
    def fromFile(cls,files,backbone='rna'):
        """Create a Multilign_object from a set of input files
        Args:
            files(str): A list of lists describing the input and output for the calculation.
                Each list should follow the format: [input file, output file, constraints,
                SHAPE]. Each value is a string with a path to the appropriate file. If no
                constraints or SHAPE information are being used, these should be empty
                strings. Input sequences should be in .seq format; output will be in .ct
                format.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        polymer_valid(backbone)
        return cls(files,polymers[backbone],None)

@e.decorate_methods(e.check_for_errors,e.not_excluded)
class Oligowalk_object(RNAstructure_wrap.Oligowalk_object):
    @e.check_for_errors
    def __init__(self,*args):
        """Initializes a Oligowalk_object using a thin wrapper around the C++ constructor (see
        RNAstructure docs for more details.) The class method Oligowalk_object.fromFile
        provides a better way to initialize Oligowalk_objects.
        To create an empty Oligowalk_object for Oligoscreen calculations, call this with no
        arguments
        """
        RNAstructure_wrap.Oligowalk_object.__init__(self,*args)

    @classmethod
    def fromFile(cls,seqfile,filetype=None):
        """Creates an Oligowalk_object from an input file.
        Args:
            seqfile(str): A path to the location of the input file
            filetype(str): The format of the input file. Supported options are "seq",
                "ct","sav" (save file from the Fold program), and "pfs" (save file
                from the partition program). If no file type is provided, this function
                will attempt to infer the filetype automatically, and raise an IOError
                if the file cannot be understood.
        """
        if filetype is None:
            filetype = interpret_filetype(seqfile)
        filetype_valid(filetype)
        return cls(seqfile,filecodes[filetype])

    @classmethod
    def fromString(cls,sequence):
        """Creates an Oligowalk_object from a sequence
        Args:
            sequence(str): The sequence for the Oligowalk_object. Input sequence should contain
                A,C,G,T,U,a,c,g,t,u,x,X. Capitalization makes no difference.
        """
        sequence_valid(sequence)
        return cls(sequence)

@e.decorate_methods(e.check_for_errors,e.not_excluded)
class ProbScan(RNAstructure_wrap.ProbScan):
    """This class is an RNA with additional functionality to calculate loop probabilities
    There are A LOT of methods available because this inherits from RNA. However, the
    following new public methods are specific to the loop probability calculation:
        probability_of_all_hairpins
        probability_of_all_internal_loops
        probability_of_all_helices
        probability_of_hairpin
        probability_of_internal_loop
        probability_of_helix
        probability_of_stack
        probability_of_multibranch_loop
    """

    @e.check_for_errors
    def __init__(self,*args):
        RNAstructure_wrap.ProbScan.__init__(self,*args)

    def __iter__(self):
        return self.iterNucs()

    def __len__(self):
        return self.GetSequenceLength()

    def __repr__(self):
        seq = "".join(nuc for nuc in self)
        return "ProbScan: %s%s" % (seq if len(self)<20 else seq[:15],
                              "" if len(self)<20 else "... %s"%seq[-4:])
    def __str__(self):
        return "".join(nuc for nuc in self)

    @classmethod
    def fromFile(cls,seqfile,filetype=None,backbone="rna"):
        """Create a ProbScan object from an input file.
        Args:
            seqfile(str): A path to the location of the input file
            filetype(str): The format of the input file. Supported options are "seq",
                and "pfs" (save file from the partition program). If no file type is
                provided, this function will attempt to infer the filetype
                automatically, and raise an IOError if the file cannot be understood.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        if filetype is None:
            filetype = interpret_filetype(seqfile)
        filetype_valid(filetype)
        polymer_valid(backbone)
        return cls(seqfile,filetype=='seq',polymers[backbone])

    @classmethod
    def fromString(cls,sequence,backbone = "rna"):
        """Create a ProbScan object from a string containin the sequence
        Args:
            sequence(str): The sequence for the RNA. Input sequence should contain
                A,C,G,T,U,a,c,g,t,u,x,X. Capitalization makes no difference. If the
                sequence is an RNA, T is treated as U.
            backbone(str,optional): The strand type, which determines which set of
                nearest neighbor parameters are used. RNA parameters are used by
                default, unless backbone is 'dna', in which case DNA parameters are used.
        """
        polymer_valid(backbone)
        sequence_valid(sequence)
        return cls(sequence,polymers[backbone])

    def probability_of_all_hairpins(self,threshold=0.01):
        """Finds all possible hairpins with probabilities greater than some threshold
        Args:
            threshold(float,optional): The minimum probability of hairpins to be
                returned. Default is 0.01
        Returns:
            A list of Hairpin objects
        """
        return map(self.Hairpin,RNAstructure_wrap.ProbScan.probability_of_all_hairpins(
            self,3,RNAstructure_wrap.ProbScan.GetSequenceLength(self),threshold))
    def probability_of_all_internal_loops(self,threshold=0.01,mode='both'):
        """Finds all possible internal and/or bulge  loops with probabilities greater than some threshold
        Args:
            threshold(float, optional): The minimum probability of loops to be returned
                Default is 0.01
            mode(str,optional): Describes what type of loops to search for. Allowed values
                are 'internal', 'bulge', and 'both'. Default is 'both'.
        Returns:
            A list of InternalLoop objects
        """
        return map(self.InternalLoop,RNAstructure_wrap.ProbScan.probability_of_all_internal_loops(self,threshold,mode))
    def probability_of_all_helices(self,threshold=0.01,length=1):
        """Finds all possible helices with probabilities greater than some threshold
        Args:
            threshold(float,optional): The minimum probability of helices to be
                returned. Default is 0.01
            length(int,optional): The size of helices to search for, in base pair STACKS
                (a helix with two base pairs has one stack). Default is 1.
        Returns:
            A list of Helix objects
        """
        return map(self.Helix,RNAstructure_wrap.ProbScan.probability_of_all_helices(self,threshold,length))
    def probability_of_multibranch_loop(self,pairs):
        """Calculates the probability of a multibranch loop, provided as a list of pairs
        Args:
            pairs(iterable): A list or tuple of pairs (any iterable with two items)
                that describe the multibranch loop. Example: [(10,40),(12,30),(31,39)]
        Returns:
            The probability of the multibranch loop, as a float
        """
        p = sorted(pairs)
        mbl = RNAstructure_wrap.multibranch_loop(p[0][0],p[0][1])
        for pair in p[1:]:
            i,j = pair
            RNAstructure_wrap.add_branch(mbl,i,j)
        return RNAstructure_wrap.ProbScan.probability_of_multibranch_loop(self,mbl)

    class Hairpin:
        """A representation of a hairpin loop.
        Attributes:
            probability(float): The probability of this loop
            i (int): The position of the 5' nucleotide closing this loop
            j (int): The position of the 3' nucleotide closing this loop
        """
        def __init__(self,hpt):
            self.probability = hpt.probability
            self.i = hpt.i
            self.j = hpt.j
        def __repr__(self):
            return "Hairpin Loop: %0.3f %d %d" % (self.probability,self.i,self.j)
        __string__ = __repr__

    class InternalLoop:
        """A representation of an internal or bulge loop
        Attributes:
            probability(float): The probability of this loop
            i (int): The position of the 5' nucleotide closing this loop on the exterior
            j (int): The position of the 3' nucleotide closing this loop on the exterior
            k (int): The position of the 5' nucleotide closing this loop on the interior
            l (int): The position of the 3' nucleotide closing this loop on the interior
        """
        def __init__(self,hpt):
            self.probability = hpt.probability
            self.i = hpt.i
            self.j = hpt.j
            self.k = hpt.k
            self.l = hpt.l
        def __repr__(self):
            return "Internal Loop: %0.3f %d %d %d %d" % (self.probability,self.i,self.j,self.k,self.l)

    class Helix:
        """A representation of an internal or bulge loop
        Attributes:
            probability(float): The probability of this helix
            i (int): The position of the 5' nucleotide closing this helix on the exterior
            j (int): The position of the 3' nucleotide closing this helix on the exterior
            k (int): The position of the 5' nucleotide closing this helix on the interior
            l (int): The position of the 3' nucleotide closing this helix on the interior
        """
        def __init__(self,hpt):
            self.probability = hpt.probability
            self.i = hpt.i
            self.j = hpt.j
            self.k = hpt.k
            self.l = hpt.l
        def __repr__(self):
            return "Helix: %0.3f %d %d %d %d" % (self.probability,self.i,self.j,self.k,self.l)

    def iterNucs(self):
        """iterate over nucleotide identities.
        note that this is has the same behavior as iterating directly
        on the ProbScan object."""
        return NucIterator(self)

    def iterIndices(self):
        """iterate over 1-indexed nucleotide indices"""
        return range(1,self.GetSequenceLength()+1)

#iterate over nucleotides for RNA and ProbScan
class NucIterator(object):
    def __init__(self,rna):
        self.i = 0
        self.imax = rna.GetSequenceLength()
        self.rna = rna

    def __iter__(self):
        return self

    def __next__(self):
        self.i += 1
        if self.i > self.imax:
            raise StopIteration
        return self.rna.GetNucleotide(self.i)

    next = __next__     # Python2 support



from inspect import signature
import timeit
import psutil
import multiprocessing
from itertools import product
import subprocess 
WINDOWSIZE = [int(i) for i in subprocess.run(
    'stty size', shell=True, stdout=subprocess.PIPE, encoding='utf-8').stdout.split()]

class FT_Decorator():
    def __init__(self,freq=1,callback=None,show=True):
        self.freq = freq
        self.callback = callback
        self.avgtime = 0
        self.count = 0
        self.show  = show

    def __call__(self,func):
        """
        decorator for printing execution time of a function.
        """
        sig = signature(func)
        def wrapped(*args, **kwargs):
            t1 = timeit.default_timer()
            result = func(*args, **kwargs)
            t2 = timeit.default_timer()


            self.avgtime = (self.count * self.avgtime + t2 - t1 ) / (self.count + 1)
            self.count += 1

            if self.callback:
                self.callback(t2-t1)

            if self.show and (self.count % self.freq == 0):
                print('Run {} {} times: avg: {:.5f}s; latest: {:.5f}s'.format(
                    func.__name__, self.count, self.avgtime, t2-t1))
            return result
        wrapped.__signature__ = sig
        wrapped.__name__ = func.__name__
        return wrapped

def ft_decorator(freq=1):
    def _ft_decorator(func):
        """
        decorator for printing execution time of a function.
        """
        sig = signature(func)
        avgtime = 0
        runcount = 0
        def wrapped(*args, **kwargs):
            nonlocal avgtime, runcount
            t1 = timeit.default_timer()
            result = func(*args, **kwargs)
            t2 = timeit.default_timer()
            avgtime = (runcount * avgtime + t2 - t1 ) / (runcount + 1)
            runcount += 1
            if runcount % freq == 0:
                print('Run {} {} times: avg: {:.5f}s; para:{}{}'.format(
                    func.__name__, runcount, avgtime, args, kwargs))
            return result
        wrapped.__signature__ = sig
        wrapped.__name__ = func.__name__
        return wrapped
    return _ft_decorator

def ft(func,args,kwargs={}, number=100):
    """
    use timeit to time [number] execution of function.
    """
    def wrapper():
        return func(*args,**kwargs)
    t = timeit.timeit(wrapper, number=number)
    print('Run {} {} times: total {:.6f}s; average {:.6f}s.'.format(func.__name__, number, t,t/number))
    return t

def poolMap(task,workload,initializer=None,initargs=None,chunks=None,
            total=None,progress_callback=None,progress_gap=(0,100),**kwargs):
    """
    speed up task in a list by multiprocessing.
    task is the function to apply on workload.
    workload is a iterable containing the task inputs.
    initializerr and initargs can be used to setup the function.
    """
    workerheads=psutil.cpu_count(logical=False)
    worker=multiprocessing.Pool(workerheads,initializer,initargs)
    total = total or len(workload)
    chunksize= int(total//chunks)+1 if chunks else int(total/workerheads/10+1)
    result = []
    count=0
    progress = progress_gap[0]
    for _ in worker.imap(task, workload, chunksize):
      count+=1
      result.append(_)
      if progress_callback :
          current_pro = count/total*(progress_gap[1]-progress_gap[0])+progress_gap[0]
          if current_pro > progress + 1:
              progress = current_pro
              progress_callback(current_pro)
    worker.close()
    worker.join()
    worker.terminate()
    return result


def cartesian(*arrays):
    """
    Give a list of arrays, return iterator of product of elements from arrays.
    similar to python product, but also yield indexes.
    """
    lenths = [len(i) for i in arrays]
    l = product(*lenths)
    p = product(*arrays)
    for i,j in zip(l,p):
        yield i,j



class MyPrint:
    """
    use mprint the same way as print().
    Set printToScreen = False to disable print. 
    Set mprint.callback to provide a callabck(msg) function when printToScreen is disabled.
    by default, call back is void().
    """
    printToScreen = True
  

    def callback(self,msg):
        return 
        
    def __call__(self,msg):
        if self.printToScreen:
            print(msg)
        else:
            self.callback(msg)
   
mprint = MyPrint()


class ProgressBar:
    """
    Class to display a progress bar, or pure count at a given frequency. 
    """
    def __init__(self,start=0,limits=None,interval=1,prefix='Progress',suffix="",decimals=2,length=None,frequency=0.1):
        """
        start: the start progress in percentage 
        interval: the start number and end number if going to use intermediate number as input. 
        prefix, suffix: text to display. 
        length: length of the bar. 
        interval: interval of nupdate if using display_number
        frequency: update frequency in percentage.
        limits: the upper and lower limit of the iteration number to use. 
        """
        if WINDOWSIZE and length == None:
            TermRow, TermCol = WINDOWSIZE
            length = int(TermCol*0.6)
        else:
            length = length or 100
        self.prefix = prefix 
        self.length = length 
        self.suffix = suffix 
        self.template = "\r{} |{}| {:." + str(decimals) + "%} {}"
        self.progress = start
        self.frequency = frequency / 100
        self.interval = interval
        self.limits = limits
        self.default = None
       

    def __call__(self,value):
        "automatically determine what to use and use that for later. "
       
        if value<1:
            return self.display_percent(value)
        else:
            return self.display_iteration(value)

    def draw_bar(self):
        filllength = int(self.length * max(min(self.progress, 1),0)) 
        bar = '█'*filllength + '-'*(self.length - filllength)
        print(self.template.format(self.prefix, bar, self.progress, self.suffix),end='\r')

    def draw_number(self):
       
        num = "{:.0f} ".format(self.progress) 
        bar = " "*(self.length - len(num)-1) + num 
        print(f"\r{self.prefix} |{bar}| {self.suffix}",end='\r')

    def display_percent(self,percent):
        "give a percentage, display if its above the frequency."
        if percent - self.progress >= self.frequency:
            self.progress = percent 
            self.draw_bar()

    def display_iteration(self,iteration):
        ""
        if not self.limits: 
            return self.display_number(iteration)
        p = (iteration - self.limits[0]) / (self.limits[1] - self.limits[0])
        return self.display_percent(p) 
    
    def display_number(self,number):
        if number - self.progress >= self.interval:
            self.progress = number
            self.draw_number()
            
    def end_bar(self):
        print()



import click
import re
from cli.utils import TableDisplay
from mymodule import mkdirs
import os
from datetime import datetime
from pathlib import PurePath
import shutil

@click.command()
@click.option('--folder','-f',default=None,type=click.Path(exists=False, file_okay=False,resolve_path=True),help="Provide folder to clean.")
@click.option('--pattern','-p',default=None, help = "Use custom regex pattern to delete.")
@click.option('--reverse','-r',is_flag=True,default=False,help='Reverse the deletion, move files and folders back.')
def cli(folder,pattern,reverse):
    """
    Clean files from a folder using regex pattern search.\n
    Default using icloud sync duplicate pattern / [2-9]$/ .
    """
    td=TableDisplay()
    op = 'reverse' if reverse else 'clean'
    folder = folder or click.prompt(td(text=f"Enter [path/to/folder] you want to {op}:")+'\n', default=os.getcwd())
    if reverse:
        try:
            click.echo(td(text=f'Reverse cleaning on folder [{folder}] ...'))
            reverse_deletion(folder)
            click.echo(td(title=">>> <g>SUCCESS</g> <<<",
                          text=f'Reversed cleaning on folder [{folder}]'))
        except Exception as e:
            click.echo(td(title=">>> <r>ERROR</r> <<<",text=f"Error: <r>{e}</r>"))
        return 
    pattern = pattern or click.prompt(td(text="Enter regex pattern:")+'\n', default=" [2-9]$") 
    try:
        pattern = re.compile(pattern)
    except re.error as e:
        click.echo(td(title=">>> <r>ERROR</r> <<<",text=f"Regex compile error: <r>{e}</r>"))
        return 
    if os.path.isdir(folder):       
        click.echo(td(text=f'Cleaning folder [{folder}] ...'))
        save , fileCount, folderCount= delete_by_pattern(folder,pattern) 
        click.echo(td(title=">>> <g>Result</g> <<<",
            text=f'Cleaning Done.\n{{{fileCount}}} files\n{{{folderCount}}} folders\nMoved to [{save}].'))
    else:
        click.echo(td(text=f"Path <r>{folder}</r> is invalid."))

def reverse_deletion(path):
    path = PurePath(path)
    tempfolder = path / '!DELETE SCRIPT TEMP FOLDER'
    filepath = tempfolder / 'Deleted files'
    folderpath = tempfolder / 'Deleted folders'
    f = open(tempfolder/'!Deletion Log.txt', 'r')
    def move(f,t,tag):
        if os.path.exists(f):
            shutil.move(f,t) 
            return
        click.echo(f'Miss [{tag}] <{filepath/p}>')

    for line in f.readlines():
        p = line.strip()[15:-1]
        if line.startswith("Move [ FILE ]"):
            move(filepath/p, path/p,' FILE ')
        elif line.startswith("Move [FOLDER]"):
            move(folderpath/p, path/p,'FOLDER')
    f.close()
   

def delete_by_pattern(path,pattern):
    regex = pattern.pattern
    movedFiles = 0
    movedFolders = 0
    tempfolder = PurePath(path) / '!DELETE SCRIPT TEMP FOLDER'
    filepath = tempfolder / 'Deleted files'
    folderpath = tempfolder / 'Deleted folders'
    mkdirs(tempfolder,filepath,folderpath)
    f = open(tempfolder/'!Deletion Log.txt', 'a')
    f.write(f'Start cleaning {path} on {datetime.now().strftime("%Y/%m/%d %H:%M")}\n')
    f.write(f"Cleaning pattern: >{regex}<\n")
    f.write('='*50+'\n')
    for root, dirs, files in os.walk(path):
        if str(tempfolder) in root:  # avoid looking into temp path.
            continue
        relative = PurePath(root).relative_to(path)
        if pattern.search(root):  # if the root folder is matched.
            click.echo(f'Move [FOLDER] <{relative}>')
            f.write(f'Move [FOLDER] <{relative}>\n')
            shutil.move(root, folderpath / relative)
            movedFolders += 1
            continue
        for file in files:
            if pattern.search(file):
                click.echo(f'Move [ FILE ] <{file}>')
                f.write(f'Move [ FILE ] <{os.path.join(relative, file)}>\n')
                os.makedirs(filepath / relative, exist_ok=True)
                shutil.move(os.path.join(root, file), filepath/relative/file)
                movedFiles += 1 #.append(oas.path.join(root, file))

    f.write('='*50+'\n')
    f.write(
        f'Cleaned {movedFiles} files, {movedFolders} folders, finished on {datetime.now().strftime("%Y/%m/%d %H:%M")}\n\n')
    f.close()
    return tempfolder, movedFiles, movedFolders



REVERSETRANS = str.maketrans("ACTGNactgn-", "TGACNtgacn-")


def revcomp(s):
    """
    return reverse complement of a sequence.
    """
    # define the dic out side is much faster.    
    # comp = map(lambda x: REVSECOMPDICT[x], s[::-1]) # slightly slower
    return s.translate(REVERSETRANS)[::-1]


# def lev_distance(s1, s2, threshold=1000):
#     """
#     calculate lev_distance of s1 and s2.
#     use threshold to calculate diagonally.stop at threshold. fastest.
#     """
#     if len(s1) > len(s2):
#         s1, s2 = s2, s1
#     l1 = len(s1)
#     vertical = [0]
#     horizontal = [0]
#     for i in range(l1):
#         char1 = s1[i]
#         char2 = s2[i]
#         newvertical = [i+1]
#         newhorizontal = [i+1]
#         for k in range(i):
#             if char1 == s2[k]:
#                 newvertical.append(vertical[k])
#             else:
#                 newvertical.append(
#                     1+min(newvertical[-1], vertical[k], vertical[k+1]))
#             if char2 == s1[k]:
#                 newhorizontal.append(horizontal[k])
#             else:
#                 newhorizontal.append(
#                     1+min(newhorizontal[-1], horizontal[k], horizontal[k+1]))
#         last = vertical[-1] if char1 == char2 else (
#             1+min(newvertical[-1], newhorizontal[-1], vertical[-1]))
#         newhorizontal.append(last)
#         newvertical.append(last)
#         currentmin = min(min(newhorizontal), min(newvertical))
#         if currentmin > threshold:
#             return currentmin
#         vertical, horizontal = newvertical, newhorizontal
#     horizontal.append(last)
#     for index2, char2 in enumerate(s2[l1:]):
#         newhorizontal = [index2+l1+1]
#         for index1, char1 in enumerate(s1):
#             if char1 == char2:
#                 newhorizontal.append(horizontal[index1])
#             else:
#                 newhorizontal.append(1 + min((horizontal[index1],
#                                               horizontal[index1 + 1],
#                                               newhorizontal[-1])))
#         currentmin = min(newhorizontal)
#         if currentmin > threshold:
#             return currentmin
#         horizontal = newhorizontal
#     return horizontal[-1]



import setuptools



with open("README.md", "r") as fh:
    long_description = fh.read()

with open("cli/_version.py","r") as f:
    _version = f.read().split('=')[1].strip(" \'\"")


setuptools.setup(
    name="huik-module", # Replace with your own username this is for PIP install, not the real python package name. the real name will be example_pkg
    version=_version,
    author="Hui Kang",
    author_email="rocksnow1942@gmail.com",
    description="Modules for my scripts.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rocksnow1942/mymodule",
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests","conf"]), # or can manually enter.
    # package_data = {'cli':['plugins/*.py']},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points = """
        [console_scripts]
        ok=cli:menu
        tl=cli:toollist
        fd=cli:folder
    """,
    install_requires=['psutil','click','colorama','terminaltables'],
    python_requires='>=3.6',
)

